﻿using System.Linq.Expressions;
using MongoDB.Driver;
using MongoDB.Driver.Linq;

namespace BPA.Component.MongoClient
{
    /// <summary>
    /// MongoDb基础仓储
    /// </summary>
    public abstract class BaseMGRepository<TDocument> : IMGRepository<TDocument> where TDocument : new()
    {
        /// <summary>
        /// 当前企业(商户)Id
        /// </summary>
        public long CorpId { protected set; get; }

        /// <summary>
        /// 当前MongoDB上下文
        /// </summary>
        public MongoDbContext CurrMongoDbContext { protected set; get; }

        /// <summary>
        /// 当前需要操作的文档集
        /// </summary>
        public IMongoCollection<TDocument> CurrentCollection { protected set; get; }

        /// <summary>
        /// MongoDbProvider
        /// </summary>
        public MongoDbProvider MongoDbProvider { protected set; get; }

        /// <summary>
        /// 构造方法
        /// </summary>
        /// <param name="mongoDbProvider"></param>
        public BaseMGRepository(MongoDbProvider mongoDbProvider)
        {
            MongoDbProvider = mongoDbProvider;
        }

        /// <summary>
        /// 切换企业Id
        /// </summary>
        /// <param name="corpId"></param>
        public virtual void ChangeCorpId(long corpId)
        {
            CorpId = corpId;
            CurrMongoDbContext = MongoDbProvider.GetDbContext(corpId);
            CurrentCollection = CurrMongoDbContext.GetCollection<TDocument>(null);
        }

        /// <summary>
        /// 检查当前文档集
        /// </summary>
        private void CheckCurrentCollection()
        {
            if (CurrentCollection == null)
                throw new ArgumentNullException($"当前要操作的文档集为空,请调用{nameof(ChangeCorpId)}方法切换企业Id!");
        }

        #region 插入扩展

        /// <summary>
        /// 插入一个 ,等同使用:<code>IMongoCollection.InsertOne</code>
        /// </summary> 
        /// <param name="document">文档对象</param>
        /// <param name="options">插入配置选项</param>
        public void InsertOne(TDocument document, InsertOneOptions? options = null)
        {
            CheckCurrentCollection();
            CurrentCollection.InsertOne(document, options);
        }

        /// <summary>
        /// 插入多个,等同使用:
        /// <code>IMongoCollection.InsertMany(IClientSessionHandle, IEnumerable{TDocument}, InsertManyOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="documents">文档对象</param>
        /// <param name="options">插入配置选项</param>
        public void InsertMany(IEnumerable<TDocument> documents, InsertManyOptions? options = null)
        {
            CheckCurrentCollection();
            CurrentCollection.InsertMany(documents, options);
        }

        /// <summary>
        /// 插入一个 ,等同使用:<code>IMongoCollection.InsertOneAsync</code>
        /// </summary> 
        /// <param name="document">文档对象</param>
        /// <param name="options">插入配置选项</param>
        public Task InsertOneAsync(TDocument document, InsertOneOptions? options = null)
        {
            CheckCurrentCollection();
            return CurrentCollection.InsertOneAsync(document, options);
        }

        /// <summary>
        /// 插入多个,等同使用:
        /// <code>IMongoCollection.InsertManyAsync(IEnumerable{TDocument}, InsertManyOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="documents">文档对象</param>
        /// <param name="options">插入配置选项</param>
        public Task InsertManyAsync(IEnumerable<TDocument> documents, InsertManyOptions? options = null)
        {
            CheckCurrentCollection();
            return CurrentCollection.InsertManyAsync(documents, options);
        }

        #endregion

        #region 更新扩展

        /// <summary>
        /// 更新满足条件的第一条数据,
        /// <code>IMongoCollection.UpdateOne(FilterDefinition{TDocument}, UpdateDefinition{TDocument}, UpdateOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="document"></param>
        /// <param name="whereExpression"></param>
        /// <param name="options"></param>
        public UpdateResult UpdateOne(TDocument document, Expression<Func<TDocument, bool>> whereExpression,
            UpdateOptions options = null)
        {
            CheckCurrentCollection();
            return CurrentCollection.UpdateOne(document, whereExpression, options);
        }

        /// <summary>
        /// 更新满足条件的第一条数据,
        /// <code>IMongoCollection.UpdateOneAsync(FilterDefinition{TDocument}, UpdateDefinition{TDocument}, UpdateOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="document"></param>
        /// <param name="whereExpression"></param>
        /// <param name="options"></param>
        public Task<UpdateResult> UpdateOneAsync(TDocument document, Expression<Func<TDocument, bool>> whereExpression,
            UpdateOptions options = null)
        {
            CheckCurrentCollection();
            return CurrentCollection.UpdateOneAsync(document, whereExpression, options);
        }

        /// <summary>
        /// 更新满足条件的多条数据, 
        /// <code>IMongoCollection.UpdateOne(FilterDefinition{TDocument}, UpdateDefinition{TDocument},UpdateOptions, CancellationToken)</code>
        /// </summary>
        /// <param name="document"></param>
        /// <param name="whereExpression"></param>
        /// <param name="options"></param>
        public UpdateResult UpdateMany(TDocument document, Expression<Func<TDocument, bool>> whereExpression,
            UpdateOptions options = null)
        {
            CheckCurrentCollection();
            return CurrentCollection.UpdateMany(document, whereExpression, options);
        }

        /// <summary>
        /// 更新满足条件的多条数据, 
        /// <code>IMongoCollection.UpdateOne(FilterDefinition{TDocument}, UpdateDefinition{TDocument},UpdateOptions, CancellationToken)</code>
        /// </summary>
        /// <param name="document"></param>
        /// <param name="whereExpression"></param>
        /// <param name="options"></param>
        public Task<UpdateResult> UpdateManyAsync(TDocument document, Expression<Func<TDocument, bool>> whereExpression,
            UpdateOptions options = null)
        {
            CheckCurrentCollection();
            return CurrentCollection.UpdateManyAsync(document, whereExpression, options);
        }

        /// <summary>
        ///   自定义字段更新 更新满足条件的第一条数据,实体键_id不允许更新
        /// </summary>
        /// <param name="whereExpression">更新条件</param>
        /// <param name="columnsExp">更新字段表达式:x=>new TDocument{ F1=1,F2=2 }</param> 
        /// <returns></returns>
        public UpdateResult UpdateOne(Expression<Action<TDocument>> columnsExp,
            Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.UpdateOne(columnsExp, whereExpression);
        }

        /// <summary>
        ///   自定义字段更新 更新满足条件的第一条数据,实体键_id不允许更新
        /// </summary>
        /// <param name="whereExpression">更新条件</param>
        /// <param name="columnsExp">更新字段表达式:x=>new TDocument{ F1=1,F2=2 }</param> 
        /// <returns></returns>
        public Task<UpdateResult> UpdateOneAsync(Expression<Action<TDocument>> columnsExp,
            Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.UpdateOneAsync(columnsExp, whereExpression);
        }

        /// <summary>
        /// 自定义更新字段和条件,更新单条符合条件的数据
        /// </summary>
        /// <param name="filter"></param>
        /// <param name="update"></param>
        /// <param name="options"></param>
        /// <returns></returns>
        public UpdateResult UpdateOne(FilterDefinition<TDocument> filter, UpdateDefinition<TDocument> update,
            UpdateOptions options = null)
        {
            CheckCurrentCollection();
            return CurrentCollection.UpdateOne(filter, update, options);
        }

        /// <summary>
        /// 自定义更新字段和条件,更新单条符合条件的数据
        /// </summary>
        /// <param name="filter"></param>
        /// <param name="update"></param>
        /// <param name="options"></param>
        /// <returns></returns>
        public Task<UpdateResult> UpdateOneAsync(FilterDefinition<TDocument> filter, UpdateDefinition<TDocument> update,
            UpdateOptions options = null)
        {
            CheckCurrentCollection();
            return CurrentCollection.UpdateOneAsync(filter, update, options);
        }


        /// <summary>
        /// 自定义更新字段和条件,更新条符合条件的数据
        /// </summary>
        /// <param name="filter"></param>
        /// <param name="update"></param>
        /// <param name="options"></param>
        /// <returns></returns>
        public UpdateResult UpdateMany(FilterDefinition<TDocument> filter, UpdateDefinition<TDocument> update,
            UpdateOptions options = null)
        {
            CheckCurrentCollection();
            return CurrentCollection.UpdateMany(filter, update, options);
        }

        /// <summary>
        /// 自定义更新字段和条件,更新条符合条件的数据
        /// </summary>
        /// <param name="filter"></param>
        /// <param name="update"></param>
        /// <param name="options"></param>
        /// <returns></returns>
        public Task<UpdateResult> UpdateManyAsync(FilterDefinition<TDocument> filter,
            UpdateDefinition<TDocument> update,
            UpdateOptions options = null)
        {
            CheckCurrentCollection();
            return CurrentCollection.UpdateManyAsync(filter, update, options);
        }

        /// <summary>
        ///   自定义字段更新 更新满足条件的多条数据,实体键_id不允许更新
        /// </summary> 
        /// <param name="whereExpression">更新条件</param>
        /// <param name="columnsExp">更新字段表达式:x=>new TDocument{ F1=1,F2=2 }</param> 
        /// <returns></returns>
        public UpdateResult UpdateMany(Expression<Action<TDocument>> columnsExp,
            Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.UpdateMany(columnsExp, whereExpression);
        }

        /// <summary>
        ///   自定义字段更新 更新满足条件的多条数据,实体键_id不允许更新
        /// </summary> 
        /// <param name="whereExpression">更新条件</param>
        /// <param name="columnsExp">更新字段表达式:x=>new TDocument{ F1=1,F2=2 }</param> 
        /// <returns></returns>
        public Task<UpdateResult> UpdateManyAsync(Expression<Action<TDocument>> columnsExp,
            Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.UpdateManyAsync(columnsExp, whereExpression);
        }


        /// <summary>
        /// 异步覆盖更新满足条件的第一条数据,主键字段名称必须是 Id,等同
        /// <code>IMongoCollection.ReplaceOne(FilterDefinition{TDocument}, TDocument, ReplaceOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="document">文档对象对象</param>
        /// <returns></returns>
        public ReplaceOneResult ReplaceOneById(TDocument document)
        {
            CheckCurrentCollection();
            return CurrentCollection.ReplaceOneById(document);
        }

        /// <summary>
        /// 异步覆盖更新满足条件的第一条数据,主键字段名称必须是 Id,等同
        /// <code>IMongoCollection.ReplaceOne(FilterDefinition{TDocument}, TDocument, ReplaceOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="document">文档对象对象</param>
        /// <returns></returns>
        public Task<ReplaceOneResult> ReplaceOneByIdAsync(TDocument document)
        {
            CheckCurrentCollection();
            return CurrentCollection.ReplaceOneByIdAsync(document);
        }

        /// <summary>
        /// 异步覆盖更新满足条件的第一条数据,主键字段名称必须是 ID,等同
        /// <code>IMongoCollection.ReplaceOne(FilterDefinition{TDocument}, TDocument, ReplaceOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="document">文档对象对象</param>
        /// <param name="whereExpression">查询添加表达式</param>
        /// <returns></returns>
        public ReplaceOneResult ReplaceOne(TDocument document, Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.ReplaceOne(document, whereExpression);
        }

        /// <summary>
        /// 异步覆盖更新满足条件的第一条数据,主键字段名称必须是 ID,等同
        /// <code>IMongoCollection.ReplaceOne(FilterDefinition{TDocument}, TDocument, ReplaceOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="document">文档对象对象</param>
        /// <param name="whereExpression">查询添加表达式</param>
        /// <returns></returns>
        public Task<ReplaceOneResult> ReplaceOneAsync(TDocument document,
            Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.ReplaceOneAsync(document, whereExpression);
        }

        #endregion

        #region 删除扩展

        /// <summary>
        /// 删除满足条件的第一条数据,等同于
        /// <code>IMongoCollection.DeleteOne(FilterDefinition{TDocument}, CancellationToken)</code>
        /// </summary> 
        /// <param name="whereExpression">删除条件表达式</param> 
        public DeleteResult DeleteOne(Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.DeleteOne(whereExpression);
        }

        /// <summary>
        /// 删除满足条件的第一条数据,等同于
        /// <code>IMongoCollection.DeleteOne(FilterDefinition{TDocument}, CancellationToken)</code>
        /// </summary> 
        /// <param name="whereExpression">删除条件表达式</param> 
        public Task<DeleteResult> DeleteOneAsync(Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.DeleteOneAsync(whereExpression);
        }

        /// <summary>
        /// 批量删除满足条件的多条数据,等同于
        /// <code>IMongoCollection.DeleteMany(FilterDefinition{TDocument}, CancellationToken)</code>
        /// </summary> 
        /// <param name="whereExpression">删除条件表达式</param> 
        public DeleteResult DeleteMany(Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.DeleteMany(whereExpression);
        }

        /// <summary>
        /// 批量删除满足条件的多条数据,等同于
        /// <code>IMongoCollection.DeleteMany(FilterDefinition{TDocument}, CancellationToken)</code>
        /// </summary> 
        /// <param name="whereExpression">删除条件表达式</param> 
        public Task<DeleteResult> DeleteManyAsync(Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.DeleteManyAsync(whereExpression);
        }

        #endregion

        #region 查询扩展

        /// <summary>
        /// 根据条件查询
        /// </summary> 
        /// <param name="filter"></param>
        /// <param name="options"></param>
        /// <returns></returns>
        public IFindFluent<TDocument, TDocument> Find(FilterDefinition<TDocument> filter, FindOptions? options)
        {
            CheckCurrentCollection();
            return CurrentCollection.Find(filter, options);
        }

        /// <summary>
        /// 根据条件查询
        /// </summary> 
        /// <param name="filter"></param>
        /// <param name="options"></param>
        /// <returns></returns>
        public Task<IAsyncCursor<TDocument>> FindAsync(
            FilterDefinition<TDocument> filter,
            FindOptions<TDocument, TDocument>? options)
        {
            CheckCurrentCollection();
            return CurrentCollection.FindAsync(filter, options);
        }

        /// <summary>
        /// 可查询 <see cref="IMongoCollectionExtensions.AsQueryable{TDocument}(IMongoCollection{TDocument}, AggregateOptions)"/>
        /// </summary> 
        /// <returns></returns>
        public IMongoQueryable<TDocument> Queryable()
        {
            CheckCurrentCollection();
            return CurrentCollection.Queryable();
        }

        /// <summary>
        ///  可查询  根据条件 
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <returns></returns>
        public IMongoQueryable<TDocument> Queryable(Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.Queryable().Where(whereExpression);
        }

        /// <summary>
        ///  可查询  根据条件 
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <param name="selectExpression"></param>
        /// <returns></returns>
        public IMongoQueryable<TResult> Queryable<TResult>(Expression<Func<TDocument, bool>> whereExpression,
            Expression<Func<TDocument, TResult>> selectExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.Queryable().Where(whereExpression).Select(selectExpression);
        }

        /// <summary>
        /// 根据条件判断是否存在
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <returns></returns>
        public bool Any(Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.Queryable().Where(whereExpression).Any();
        }

        /// <summary>
        /// 根据ID获取entity
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        public TDocument GetById(dynamic id)
        {
            CheckCurrentCollection();
            FilterDefinition<TDocument> filter = Builders<TDocument>.Filter.Eq("_id", id);
            var find = CurrentCollection.Find(filter, null);
            return find.FirstOrDefault();
        }

        /// <summary>
        /// 根据ID获取entity
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        public async Task<TDocument> GetByIdAsync(dynamic id)
        {
            CheckCurrentCollection();
            FilterDefinition<TDocument> filter = Builders<TDocument>.Filter.Eq("_id", id);
            var find = await CurrentCollection.FindAsync(filter, null);
            return find.FirstOrDefault();
        }

        /// <summary>
        /// 根据ID获取entity
        /// </summary>
        /// <param name="id"></param>
        /// <param name="selectExpression">需要返会的字段表达式</param>
        /// <returns></returns>
        public TResult GetById<TResult>(dynamic id, Expression<Func<TDocument, TResult>> selectExpression)
        {
            CheckCurrentCollection();
            FilterDefinition<TDocument> filter = Builders<TDocument>.Filter.Eq("_id", id);
            var find = CurrentCollection.Find(filter, null);
            return find.Select(selectExpression).FirstOrDefault();
        }

        /// <summary>
        /// 根据条件 获取第一条数据
        /// </summary>
        /// <param name="whereExpression">条件</param>
        /// <param name="selectExpression">返回的新对象选择表达式</param>
        /// <returns></returns>
        public TResult? First<TResult>(Expression<Func<TDocument, bool>> whereExpression,
            Expression<Func<TDocument, TResult>> selectExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.Queryable().Where(whereExpression).Select(selectExpression).FirstOrDefault();
        }

        /// <summary>
        /// 根据条件 获取第一条数据
        /// </summary>
        /// <param name="whereExpression">条件</param>
        /// <param name="selectExpression">返回的新对象选择表达式</param>
        /// <returns></returns>
        public Task<TResult> FirstAsync<TResult>(Expression<Func<TDocument, bool>> whereExpression,
            Expression<Func<TDocument, TResult>> selectExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.Queryable().Where(whereExpression).Select(selectExpression).FirstOrDefaultAsync();
        }

        /// <summary>
        /// 根据条件 获取第一条数据
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <returns></returns>
        public TDocument? First(Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.Queryable().Where(whereExpression).FirstOrDefault();
        }

        /// <summary>
        /// 根据条件 获取第一条数据
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <returns></returns>
        public Task<TDocument> FirstAsync(Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.Queryable().Where(whereExpression).FirstOrDefaultAsync();
        }

        /// <summary>
        ///  根据条件获取集合
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <returns></returns>
        public List<TDocument> QueryList(Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.Queryable().Where(whereExpression).ToList();
        }

        /// <summary>
        ///  根据条件获取集合
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <returns></returns>
        public Task<List<TDocument>> QueryListAsync(Expression<Func<TDocument, bool>> whereExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.Queryable().Where(whereExpression).ToListAsync();
        }

        /// <summary>
        ///  根据条件获取集合
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <param name="selectExpression"></param>
        /// <returns></returns>
        public List<TResult> QueryList<TResult>(Expression<Func<TDocument, bool>> whereExpression,
            Expression<Func<TDocument, TResult>> selectExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.Queryable().Where(whereExpression).Select(selectExpression).ToList();
        }

        /// <summary>
        ///  根据条件获取集合
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <param name="selectExpression"></param>
        /// <returns></returns>
        public Task<List<TResult>> QueryListAsync<TResult>(Expression<Func<TDocument, bool>> whereExpression,
            Expression<Func<TDocument, TResult>> selectExpression)
        {
            CheckCurrentCollection();
            return CurrentCollection.Queryable().Where(whereExpression).Select(selectExpression).ToListAsync();
        }

        #endregion

        #region 创建 查询 、索引、更新、排序 条件定义

        /// <summary>
        /// 创建过滤条件定义
        /// </summary>
        public FilterDefinitionBuilder<TDocument> BuilderFilter() => Builders<TDocument>.Filter;

        /// <summary>
        /// 创建索引条件定义
        /// </summary>
        public IndexKeysDefinitionBuilder<TDocument> BuilderIndexKeys() => Builders<TDocument>.IndexKeys;

        /// <summary>
        /// 创建更新条件定义
        /// </summary>
        public UpdateDefinitionBuilder<TDocument> BuilderUpdate() => Builders<TDocument>.Update;

        /// <summary>
        /// 创建排序条件定义
        /// </summary>
        public SortDefinitionBuilder<TDocument> BuilderSort() => Builders<TDocument>.Sort;

        /// <summary>
        /// Filter
        /// </summary>
        public FilterDefinitionBuilder<TDocument> Filter => Builders<TDocument>.Filter;

        /// <summary>
        /// Update
        /// </summary>
        public UpdateDefinitionBuilder<TDocument> Update => Builders<TDocument>.Update;

        /// <summary>
        /// Sort
        /// </summary>
        public SortDefinitionBuilder<TDocument> Sort => Builders<TDocument>.Sort;

        #endregion
    }
}