﻿using System.Linq.Expressions;
using MongoDB.Driver;
using MongoDB.Driver.Linq;

namespace BPA.Component.MongoClient
{
    /// <summary>
    /// MongoDB仓储接口
    /// </summary>
    public interface IMGRepository<TDocument> where TDocument : new()
    {
        /// <summary>
        /// 当前企业(商户)Id
        /// </summary>
        long CorpId { get; }

        /// <summary>
        /// 切换企业Id
        /// </summary>
        /// <param name="corpId"></param>
        void ChangeCorpId(long corpId);

        /// <summary>
        ///  MongonDbProvider
        /// </summary>
        MongoDbProvider MongoDbProvider { get; }

        /// <summary>
        /// 当前MongoDB上下文
        /// </summary>
        MongoDbContext CurrMongoDbContext { get; }

        /// <summary>
        /// 当前需要操作的文档集
        /// </summary>
        IMongoCollection<TDocument> CurrentCollection { get; }

        #region 插入扩展

        /// <summary>
        /// 插入一个 ,等同使用:<code>IMongoCollection.InsertOne</code>
        /// </summary>
        /// <param name="document">文档对象</param>
        /// <param name="options">插入配置选项</param>
        void InsertOne(TDocument document, InsertOneOptions? options = null);

        /// <summary>
        /// 插入多个,等同使用:
        /// <code>IMongoCollection.InsertMany(IClientSessionHandle, IEnumerable{TDocument}, InsertManyOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="documents">文档对象</param>
        /// <param name="options">插入配置选项</param>
        void InsertMany(IEnumerable<TDocument> documents, InsertManyOptions? options = null);

        /// <summary>
        /// 插入一个 ,等同使用:<code>IMongoCollection.InsertOneAsync</code>
        /// </summary> 
        /// <param name="document">文档对象</param>
        /// <param name="options">插入配置选项</param>
        Task InsertOneAsync(TDocument document, InsertOneOptions options = null);

        /// <summary>
        /// 插入多个,等同使用:
        /// <code>IMongoCollection.InsertManyAsync(IEnumerable{TDocument}, InsertManyOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="documents">文档对象</param>
        /// <param name="options">插入配置选项</param>
        Task InsertManyAsync(IEnumerable<TDocument> documents, InsertManyOptions options = null);

        #endregion

        #region 更新扩展

        /// <summary>
        /// 更新满足条件的第一条数据,
        /// <code>IMongoCollection.UpdateOne(FilterDefinition{TDocument}, UpdateDefinition{TDocument}, UpdateOptions, CancellationToken)</code>
        /// </summary>
        /// <param name="document"></param>
        /// <param name="whereExpression"></param>
        /// <param name="options"></param>
        UpdateResult UpdateOne(TDocument document, Expression<Func<TDocument, bool>> whereExpression, UpdateOptions options = null);

        /// <summary>
        /// 更新满足条件的第一条数据,
        /// <code>IMongoCollection.UpdateOneAsync(FilterDefinition{TDocument}, UpdateDefinition{TDocument}, UpdateOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="document"></param>
        /// <param name="whereExpression"></param>
        /// <param name="options"></param>
        Task<UpdateResult> UpdateOneAsync(TDocument document, Expression<Func<TDocument, bool>> whereExpression,
            UpdateOptions options = null);

        /// <summary>
        /// 更新满足条件的多条数据, 
        /// <code>IMongoCollection.UpdateOne(FilterDefinition{TDocument}, UpdateDefinition{TDocument},UpdateOptions, CancellationToken)</code>
        /// </summary>
        /// <param name="document"></param>
        /// <param name="whereExpression"></param>
        /// <param name="options"></param>
        UpdateResult UpdateMany(TDocument document, Expression<Func<TDocument, bool>> whereExpression, UpdateOptions options = null);

        /// <summary>
        /// 更新满足条件的多条数据, 
        /// <code>IMongoCollection.UpdateOne(FilterDefinition{TDocument}, UpdateDefinition{TDocument},UpdateOptions, CancellationToken)</code>
        /// </summary>
        /// <param name="document"></param>
        /// <param name="whereExpression"></param>
        /// <param name="options"></param>
        Task<UpdateResult> UpdateManyAsync(TDocument document, Expression<Func<TDocument, bool>> whereExpression,
            UpdateOptions options = null);

        /// <summary>
        ///   自定义字段更新 更新满足条件的第一条数据,实体键_id不允许更新
        /// </summary>
        /// <param name="whereExpression">更新条件</param>
        /// <param name="columnsExp">更新字段表达式:x=>new TDocument{ F1=1,F2=2 }</param> 
        /// <returns></returns>
        UpdateResult UpdateOne(Expression<Action<TDocument>> columnsExp, Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        ///   自定义字段更新 更新满足条件的第一条数据,实体键_id不允许更新
        /// </summary>
        /// <param name="whereExpression">更新条件</param>
        /// <param name="columnsExp">更新字段表达式:x=>new TDocument{ F1=1,F2=2 }</param> 
        /// <returns></returns>
        Task<UpdateResult> UpdateOneAsync(Expression<Action<TDocument>> columnsExp, Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        ///   自定义字段更新 更新满足条件的多条数据,实体键_id不允许更新
        /// </summary>
        /// <param name="whereExpression">更新条件</param>
        /// <param name="columnsExp">更新字段表达式:x=>new TDocument{ F1=1,F2=2 }</param> 
        /// <returns></returns>
        UpdateResult UpdateMany(Expression<Action<TDocument>> columnsExp, Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        ///   自定义字段更新 更新满足条件的多条数据,实体键_id不允许更新
        /// </summary> 
        /// <param name="whereExpression">更新条件</param>
        /// <param name="columnsExp">更新字段表达式:x=>new TDocument{ F1=1,F2=2 }</param> 
        /// <returns></returns>
        Task<UpdateResult> UpdateManyAsync(Expression<Action<TDocument>> columnsExp, Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        /// 自定义更新字段和条件,更新单条符合条件的数据
        /// </summary>
        /// <param name="filter"></param>
        /// <param name="update"></param>
        /// <param name="options"></param>
        /// <returns></returns>
        UpdateResult UpdateOne(FilterDefinition<TDocument> filter, UpdateDefinition<TDocument> update, UpdateOptions options = null);

        /// <summary>
        /// 自定义更新字段和条件,更新单条符合条件的数据
        /// </summary>
        /// <param name="filter"></param>
        /// <param name="update"></param>
        /// <param name="options"></param>
        /// <returns></returns>
        Task<UpdateResult> UpdateOneAsync(FilterDefinition<TDocument> filter, UpdateDefinition<TDocument> update,
            UpdateOptions options = null);

        /// <summary>
        /// 自定义更新字段和条件,更新条符合条件的数据
        /// </summary>
        /// <param name="filter"></param>
        /// <param name="update"></param>
        /// <param name="options"></param>
        /// <returns></returns>
        UpdateResult UpdateMany(FilterDefinition<TDocument> filter, UpdateDefinition<TDocument> update, UpdateOptions options = null);

        /// <summary>
        /// 自定义更新字段和条件,更新条符合条件的数据
        /// </summary>
        /// <param name="filter"></param>
        /// <param name="update"></param>
        /// <param name="options"></param>
        /// <returns></returns>
        Task<UpdateResult> UpdateManyAsync(FilterDefinition<TDocument> filter, UpdateDefinition<TDocument> update,
            UpdateOptions options = null);

        /// <summary>
        /// 异步覆盖更新满足条件的第一条数据,主键字段名称必须是 Id,等同
        /// <code>IMongoCollection.ReplaceOne(FilterDefinition{TDocument}, TDocument, ReplaceOptions, CancellationToken)</code>
        /// </summary>
        /// <param name="document">文档对象对象</param>
        /// <returns></returns>
        ReplaceOneResult ReplaceOneById(TDocument document);


        /// <summary>
        /// 异步覆盖更新满足条件的第一条数据,主键字段名称必须是 Id,等同
        /// <code>IMongoCollection.ReplaceOne(FilterDefinition{TDocument}, TDocument, ReplaceOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="document">文档对象对象</param>
        /// <returns></returns>
        Task<ReplaceOneResult> ReplaceOneByIdAsync(TDocument document);

        /// <summary>
        /// 异步覆盖更新满足条件的第一条数据,主键字段名称必须是 ID,等同
        /// <code>IMongoCollection.ReplaceOne(FilterDefinition{TDocument}, TDocument, ReplaceOptions, CancellationToken)</code>
        /// </summary>
        /// <param name="document">文档对象对象</param>
        /// <param name="whereExpression">查询添加表达式</param>
        /// <returns></returns>
        ReplaceOneResult ReplaceOne(TDocument document, Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        /// 异步覆盖更新满足条件的第一条数据,主键字段名称必须是 ID,等同
        /// <code>IMongoCollection.ReplaceOne(FilterDefinition{TDocument}, TDocument, ReplaceOptions, CancellationToken)</code>
        /// </summary> 
        /// <param name="document">文档对象对象</param>
        /// <param name="whereExpression">查询添加表达式</param>
        /// <returns></returns>
        Task<ReplaceOneResult> ReplaceOneAsync(TDocument document, Expression<Func<TDocument, bool>> whereExpression);

        #endregion

        #region 删除扩展

        /// <summary>
        /// 删除满足条件的第一条数据,等同于
        /// <code>IMongoCollection.DeleteOne(FilterDefinition{TDocument}, CancellationToken)</code>
        /// </summary>
        /// <param name="whereExpression">删除条件表达式</param> 
        DeleteResult DeleteOne(Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        /// 删除满足条件的第一条数据,等同于
        /// <code>IMongoCollection.DeleteOne(FilterDefinition{TDocument}, CancellationToken)</code>
        /// </summary> 
        /// <param name="whereExpression">删除条件表达式</param> 
        Task<DeleteResult> DeleteOneAsync(Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        /// 批量删除满足条件的多条数据,等同于
        /// <code>IMongoCollection.DeleteMany(FilterDefinition{TDocument}, CancellationToken)</code>
        /// </summary>
        /// <param name="whereExpression">删除条件表达式</param> 
        DeleteResult DeleteMany(Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        /// 批量删除满足条件的多条数据,等同于
        /// <code>IMongoCollection.DeleteMany(FilterDefinition{TDocument}, CancellationToken)</code>
        /// </summary> 
        /// <param name="whereExpression">删除条件表达式</param> 
        Task<DeleteResult> DeleteManyAsync(Expression<Func<TDocument, bool>> whereExpression);

        #endregion

        #region 查询扩展

        /// <summary>
        /// 根据条件查询
        /// </summary> 
        /// <param name="filter"></param>
        /// <param name="options"></param>
        /// <returns></returns>
        IFindFluent<TDocument, TDocument> Find(FilterDefinition<TDocument> filter, FindOptions options = null);

        /// <summary>
        /// 根据条件查询
        /// </summary> 
        /// <param name="filter"></param>
        /// <param name="options"></param>
        /// <returns></returns>
        Task<IAsyncCursor<TDocument>> FindAsync(FilterDefinition<TDocument> filter, FindOptions<TDocument, TDocument> options = null);

        /// <summary>
        /// 可查询 <see cref="IMongoCollectionExtensions.AsQueryable{TDocument}(IMongoCollection{TDocument}, AggregateOptions)"/>
        ///   , 同步时用ToList, 异步时用 ToListAsync
        /// </summary>
        /// <returns></returns>
        IMongoQueryable<TDocument> Queryable();

        /// <summary>
        /// 可查询  根据条件   , 同步时用ToList， 异步时用 ToListAsync
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <returns></returns>
        IMongoQueryable<TDocument> Queryable(Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        ///  可查询  根据条件  , 同步时用ToList , 异步时用 ToListAsync
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <param name="selectExpression"></param>
        /// <returns></returns>
        IMongoQueryable<TResult> Queryable<TResult>(Expression<Func<TDocument, bool>> whereExpression,
            Expression<Func<TDocument, TResult>> selectExpression);

        /// <summary>
        /// 根据条件判断是否存在
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <returns></returns>
        bool Any(Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        /// 根据ID获取entity
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        TDocument GetById(dynamic id);

        /// <summary>
        /// 根据ID获取entity
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        Task<TDocument> GetByIdAsync(dynamic id);

        /// <summary>
        /// 根据ID获取entity
        /// </summary>
        /// <param name="id"></param>
        /// <param name="selectExpression">需要返回的字段表达式</param>
        /// <returns></returns>
        TResult GetById<TResult>(dynamic id, Expression<Func<TDocument, TResult>> selectExpression);

        /// <summary>
        /// 根据条件 获取第一条数据
        /// </summary>
        /// <param name="whereExpression">条件</param>
        /// <param name="selectExpression">返回的新对象选择表达式</param>
        /// <returns></returns>
        TResult? First<TResult>(Expression<Func<TDocument, bool>> whereExpression, Expression<Func<TDocument, TResult>> selectExpression);

        /// <summary>
        /// 根据条件 获取第一条数据
        /// </summary>
        /// <param name="whereExpression">条件</param>
        /// <param name="selectExpression">返回的新对象选择表达式</param>
        /// <returns></returns>
        Task<TResult> FirstAsync<TResult>(Expression<Func<TDocument, bool>> whereExpression,
            Expression<Func<TDocument, TResult>> selectExpression);

        /// <summary>
        /// 根据条件 获取第一条数据
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <returns></returns>
        TDocument? First(Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        /// 根据条件 获取第一条数据
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <returns></returns>
        Task<TDocument> FirstAsync(Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        ///  根据条件获取集合
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <returns></returns>
        List<TDocument> QueryList(Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        ///  根据条件获取集合
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <returns></returns>
        Task<List<TDocument>> QueryListAsync(Expression<Func<TDocument, bool>> whereExpression);

        /// <summary>
        ///  根据条件获取集合
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <param name="selectExpression"></param>
        /// <returns></returns>
        List<TResult> QueryList<TResult>(Expression<Func<TDocument, bool>> whereExpression,
            Expression<Func<TDocument, TResult>> selectExpression);

        /// <summary>
        ///  根据条件获取集合
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <param name="selectExpression"></param>
        /// <returns></returns>
        Task<List<TResult>> QueryListAsync<TResult>(Expression<Func<TDocument, bool>> whereExpression,
            Expression<Func<TDocument, TResult>> selectExpression);

        #endregion

        #region 创建 查询 、索引、更新、排序 条件定义

        /// <summary>
        /// Filter
        /// </summary>
        FilterDefinitionBuilder<TDocument> Filter { get; }

        /// <summary>
        /// Update
        /// </summary>
        UpdateDefinitionBuilder<TDocument> Update { get; }

        /// <summary>
        /// Sort
        /// </summary>
        SortDefinitionBuilder<TDocument> Sort { get; }

        /// <summary>
        /// 创建过滤条件定义
        /// </summary>
        FilterDefinitionBuilder<TDocument> BuilderFilter();

        /// <summary>
        /// 创建索引条件定义
        /// </summary> 
        IndexKeysDefinitionBuilder<TDocument> BuilderIndexKeys();

        /// <summary>
        /// 创建更新条件定义
        /// </summary> 
        UpdateDefinitionBuilder<TDocument> BuilderUpdate();

        /// <summary>
        /// 创建排序条件定义
        /// </summary> 
        SortDefinitionBuilder<TDocument> BuilderSort();

        #endregion
    }
}