﻿using BPA.Component.MongoClient.Conventions;
using MongoDB.Bson;
using MongoDB.Bson.Serialization;
using MongoDB.Bson.Serialization.Conventions;
using MongoDB.Bson.Serialization.Serializers;
using MongoDB.Driver;
using Newtonsoft.Json.Linq;

namespace BPA.Component.MongoClient
{
    /// <summary>
    /// MongoDbContext
    /// </summary>
    public class MongoDbContext : IDisposable
    {
        static MongoDbContext()
        {
            var serializer = new DateTimeSerializer(DateTimeKind.Local, BsonType.DateTime);
            BsonSerializer.RegisterSerializer(typeof(DateTime), serializer);
            BsonSerializer.RegisterSerializer(typeof(JObject), new JObjectSerializer());
            BsonSerializer.RegisterSerializer(typeof(JArray), new JArraySerializer());
            BsonSerializer.RegisterSerializer(typeof(Dictionary<string, object>), new DictionarySerializerSerializer());
            BsonSerializer.RegisterSerializer(typeof(decimal), new DecimalSerializer(BsonType.Decimal128));
            BsonSerializer.RegisterSerializer(typeof(decimal?),
                new NullableSerializer<decimal>(new DecimalSerializer(BsonType.Decimal128)));
            BsonSerializer.RegisterSerializationProvider(new BPASerializationProvider());

            ConventionRegistry.Register("IgnoreExtraElements",
                new ConventionPack
                {
                    new IgnoreExtraElementsConvention(true), new BPACamelCaseElementNameConvention(),
                    new EnumRepresentationConvention(BsonType.String)
                }, _ => true);
        }

        /// <summary>
        /// 当前连接
        /// </summary>
        public readonly IMongoClient MongoClient;

        /// <summary>
        /// 当前操作数据库
        /// </summary>
        public readonly IMongoDatabase Database;

        /// <summary>
        /// 企业ID
        /// </summary>
        public readonly long CorpId;

        /// <summary>
        /// 构造方法
        /// </summary>
        /// <param name="corpId"></param>
        /// <param name="mongoClient"></param>
        /// <param name="database"></param>
        public MongoDbContext(long corpId, IMongoClient mongoClient, IMongoDatabase database)
        {
            CorpId = corpId;
            MongoClient = mongoClient;
            Database = database;
        }

        /// <summary>
        /// 检查数据库是否存在
        /// </summary>
        /// <param name="mongoClient"></param>
        /// <param name="databaseName"></param>
        /// <returns></returns>
        public bool DatabaseExists(MongoDB.Driver.MongoClient mongoClient, string databaseName)
        {
            try
            {
                var dbNames = mongoClient.ListDatabases().ToList().Select(db => db.GetValue("name").AsString);
                return dbNames.Contains(databaseName);
            }
            catch
            {
                return true;
            }
        }

        /// <summary>
        /// 检查文档集是否存在
        /// </summary>
        /// <param name="mongoDatabase"></param>
        /// <param name="collectionName"></param>
        /// <returns></returns>
        public bool CollectionExists(IMongoDatabase mongoDatabase, string collectionName)
        {
            var options = new ListCollectionsOptions
            {
                Filter = Builders<BsonDocument>.Filter.Eq("name", collectionName)
            };
            return mongoDatabase.ListCollections(options).ToEnumerable().Any();
        }

        /// <summary>
        /// 获取文档集接口
        /// </summary>
        /// <typeparam name="TDocument"></typeparam>
        /// <param name="collectionName"></param>
        /// <param name="settings"></param>
        /// <returns></returns>
        public IMongoCollection<TDocument> GetCollection<TDocument>(string collectionName,
            MongoCollectionSettings? settings)
        {
            return Database.GetCollection<TDocument>(collectionName, settings);
        }


        /// <summary>
        /// 根据企业名称和文档集类型构建 文档集名称
        /// <code>
        /// 返回文档集格式:{typeof(TDocument).Name}{corpId}, 
        /// </code>
        /// </summary>
        /// <typeparam name="TDocument"></typeparam> 
        /// <param name="corpId">企业ID</param>
        /// <returns></returns>
        public string BuildCollectionName<TDocument>(long corpId)
        {
            return $"{typeof(TDocument).Name}{corpId}";
        }

        /// <summary>
        /// 根据企业ID获取，获取文档集接口
        /// </summary>
        /// <typeparam name="TDocument"></typeparam> 
        /// <param name="settings"></param>
        /// <returns></returns>
        public IMongoCollection<TDocument> GetCollection<TDocument>(MongoCollectionSettings? settings)
        {
            var collectionName = BuildCollectionName<TDocument>(CorpId);
            return GetCollection<TDocument>(collectionName, settings);
        }


        /// <summary>
        /// 检查文档集是否存在
        /// </summary>
        /// <param name="collectionName">文档集名称</param>
        /// <returns></returns>
        public bool CollectionExists(string collectionName)
        {
            var options = new ListCollectionsOptions
            {
                Filter = Builders<BsonDocument>.Filter.Eq("name", collectionName)
            };
            return Database.ListCollections(options).ToEnumerable().Any();
        }

        /// <summary>
        /// 清空文档集数据
        /// </summary>
        /// <typeparam name="TDocument"></typeparam>
        public IMongoCollection<TDocument> ClearCollection<TDocument>()
        {
            var collectionName = BuildCollectionName<TDocument>(CorpId);
            //删除文档集
            Database.DropCollection(collectionName);
            if (!CollectionExists(Database, collectionName))
            {
                this.CreateCollection<TDocument>(collectionName);
            }

            return Database.GetCollection<TDocument>(collectionName);
        }

        /// <summary>
        /// 销毁
        /// </summary>
        public void Dispose()
        {
        }
    }
}