﻿using MongoDB.Driver;

namespace BPA.Component.MongoClient
{
    /// <summary>
    /// MongoDbContext 工厂
    /// </summary> 
    public abstract class MongoDbProvider
    {
        /// <summary>
        /// 企业和别名的关系
        /// </summary>
        protected readonly Dictionary<long, MongoDbContext> CorpWithMongoDbContexts = new();

        /// <summary>
        /// 别名和mongondb实例的对应关系
        /// </summary>
        protected readonly Dictionary<string, IMongoDatabase> AliasWithMongoDbContexts = new();

        /// <summary>
        /// 别名和连接配置关系
        /// </summary>
        public IReadOnlyDictionary<string, MongoDbConfig> AliasConfig { private set; get; }

        /// <summary>
        /// MongonDbFactory
        /// </summary>
        /// <param name="mongoDbConfigs"></param>
        public MongoDbProvider(List<MongoDbConfig> mongoDbConfigs)
        {
            SetMongoDbConfigs(mongoDbConfigs);
        }

        /// <summary>
        /// 设置数据库连接配置
        /// </summary>
        /// <param name="mongoDbConfigs"></param>
        public void SetMongoDbConfigs(List<MongoDbConfig> mongoDbConfigs)
        {
            lock (this)
            {
                var dic = new Dictionary<string, MongoDbConfig>();
                foreach (var config in mongoDbConfigs)
                {
                    if (dic.ContainsKey(config.Alias))
                        throw new ArgumentException($"数据库{this.GetType().Name}的配置别名[{config.Alias}]出现重复");
                    dic.Add(config.Alias, config);
                }

                AliasConfig = dic;
            }
        }

        /// <summary>
        /// 根据企业ID获取MongoDb上下文对象
        /// </summary>
        /// <param name="corpId">企业ID</param>
        /// <returns></returns>
        public virtual MongoDbContext GetDbContext(long corpId)
        {
            if (CorpWithMongoDbContexts.ContainsKey(corpId))
                return CorpWithMongoDbContexts[corpId];
            lock (CorpWithMongoDbContexts)
            {
                if (CorpWithMongoDbContexts.ContainsKey(corpId))
                    return CorpWithMongoDbContexts[corpId];
                var dbAlias = GetDbAliasByCropId(corpId);
                if (dbAlias == null)
                    throw new ArgumentException($"无法获取到企业[{corpId}]的{this.GetType().Name}数据库别名!");
                if (!AliasConfig.ContainsKey(dbAlias))
                    throw new ArgumentException($"无法获取到企业[{corpId}]的{this.GetType().Name}数据库别名[{dbAlias}]对应的数据库连接!");
                var databases = GetMongoDatabase(dbAlias);
                return CorpWithMongoDbContexts[corpId] = new MongoDbContext(corpId, databases.Client, databases);
            }
        }

        /// <summary>
        /// GetMongoDatabase
        /// </summary>
        /// <param name="dbAlias"></param>
        /// <returns></returns>
        public IMongoDatabase GetMongoDatabase(string dbAlias)
        {
            IMongoDatabase mongoDatabase = null;
            lock (AliasWithMongoDbContexts)
            {
                if (AliasWithMongoDbContexts.ContainsKey(dbAlias))
                    mongoDatabase = AliasWithMongoDbContexts[dbAlias];
                else
                {
                    var mongoDbConfig = AliasConfig[dbAlias];
                    var client = new MongoDB.Driver.MongoClient(mongoDbConfig.ConnectionString);
                    mongoDatabase = client.GetDatabase(mongoDbConfig.DatabasesName);
                    AliasWithMongoDbContexts[dbAlias] = mongoDatabase;
                }
            }

            return mongoDatabase;
        }

        /// <summary>
        /// 获取所有的数据库操作对象  别名+数据库操作对象
        /// </summary>
        /// <returns></returns>
        public IReadOnlyDictionary<string, IMongoDatabase> GetMongoDatabases()
        {
            if (AliasConfig.Count != AliasWithMongoDbContexts.Count)
            {
                foreach (var dbAlias in AliasConfig.Keys)
                {
                    GetMongoDatabase(dbAlias);
                }
            }

            return AliasWithMongoDbContexts;
        }

        /// <summary>
        /// 根据企业Id获取数据库别名,子类必须实现
        /// </summary>
        /// <returns></returns>
        protected abstract string GetDbAliasByCropId(long corpId);

        /// <summary>
        /// 获取文档集接口
        /// </summary>
        /// <typeparam name="TDocument"></typeparam>
        /// <param name="cropId">企业ID</param> 
        /// <param name="settings"></param>
        /// <returns></returns>
        public IMongoCollection<TDocument> GetCollection<TDocument>(long cropId,
            MongoCollectionSettings settings = null)
        {
            MongoDbContext dbContext = GetDbContext(cropId);
            return dbContext.GetCollection<TDocument>(settings);
        }
    }
}